/*
 * Copyright 2012 Francesco Cappelli
 *
 * This file is part of Pagliaio.
 *
 * Pagliaio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pagliaio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pagliaio.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lucene++/LuceneHeaders.h>
#include <lucene++/WordlistLoader.h>

#include "PagliaioStandardAnalyzer.h"
#include "PagliaioStandardAnalyzerSavedStreams.h"

using namespace Lucene;

namespace pagliaio {

DECLARE_SHARED_PTR(PagliaioStandardAnalyzerSavedStreams)

PagliaioStandardAnalyzer::PagliaioStandardAnalyzer(LuceneVersion::Version matchVersion)
{
	ConstructAnalyser(matchVersion, StopAnalyzer::ENGLISH_STOP_WORDS_SET());
}

PagliaioStandardAnalyzer::PagliaioStandardAnalyzer(LuceneVersion::Version matchVersion, HashSet<String> stopWords)
{
	ConstructAnalyser(matchVersion, stopWords);
}

PagliaioStandardAnalyzer::PagliaioStandardAnalyzer(LuceneVersion::Version matchVersion, const String& stopwords)
{
	ConstructAnalyser(matchVersion, WordlistLoader::getWordSet(stopwords));
}

PagliaioStandardAnalyzer::PagliaioStandardAnalyzer(LuceneVersion::Version matchVersion, ReaderPtr stopwords)
{
	ConstructAnalyser(matchVersion, WordlistLoader::getWordSet(stopwords));
}

PagliaioStandardAnalyzer::~PagliaioStandardAnalyzer()
{
}

void PagliaioStandardAnalyzer::ConstructAnalyser(LuceneVersion::Version matchVersion, HashSet<String> stopWords)
{
	stopSet = stopWords;
	enableStopPositionIncrements = StopFilter::getEnablePositionIncrementsVersionDefault(matchVersion);
	replaceInvalidAcronym = LuceneVersion::onOrAfter(matchVersion, LuceneVersion::LUCENE_24);
	this->matchVersion = matchVersion;
}

TokenStreamPtr PagliaioStandardAnalyzer::tokenStream(const String& fieldName, ReaderPtr reader)
{
	WhitespaceTokenizerPtr tokenStream(newLucene<WhitespaceTokenizer>(reader));
	TokenStreamPtr result(newLucene<StandardFilter>(tokenStream));
	result = newLucene<LowerCaseFilter>(result);
	result = newLucene<StopFilter>(enableStopPositionIncrements, result, stopSet);
	return result;
}

TokenStreamPtr PagliaioStandardAnalyzer::reusableTokenStream(const String& fieldName, ReaderPtr reader)
{
	PagliaioStandardAnalyzerSavedStreamsPtr streams = boost::dynamic_pointer_cast<PagliaioStandardAnalyzerSavedStreams>(getPreviousTokenStream());
	if (!streams)
	{
		streams = newLucene<PagliaioStandardAnalyzerSavedStreams>();
		setPreviousTokenStream(streams);
		streams->tokenStream = newLucene<WhitespaceTokenizer>(reader);
		streams->filteredTokenStream = newLucene<StandardFilter>(streams->tokenStream);
		streams->filteredTokenStream = newLucene<LowerCaseFilter>(streams->filteredTokenStream);
		streams->filteredTokenStream = newLucene<StopFilter>(enableStopPositionIncrements, streams->filteredTokenStream, stopSet);
	}
	else
		streams->tokenStream->reset(reader);

	return streams->filteredTokenStream;
}

PagliaioStandardAnalyzerSavedStreams::~PagliaioStandardAnalyzerSavedStreams()
{
}

}
