/*
 * Copyright 2012 Francesco Cappelli
 *
 * This file is part of Pagliaio.
 *
 * Pagliaio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pagliaio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pagliaio.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <csignal>

#include <lucene++/LuceneHeaders.h>
#include <lucene++/FileUtils.h>

extern "C" {
#include <sys/time.h>
#include <stdint.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <microhttpd.h>
#include <syslog.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
}

#include <boost/thread.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <version.h>
#include <utils.h>

#include <Configuration.h>

using namespace std;

Lucene::IndexWriterPtr indexWriter;
Lucene::IndexReaderPtr indexReader;
Lucene::SearcherPtr indexSearcher;

boost::shared_mutex indexMutex;

Lucene::String directoryFolder = L"/var/run/pagliaio";
unsigned int httpPort = 4200;
char* bindAddr = (char*)"0.0.0.0";
unsigned int threadSleepSeconds = 5;

char* configurationFilePath = (char*)"/etc/pagliaio/pagliaio.xml";
pagliaio::Configuration configuration;

bool daemonRunning = true;

int httpRequestsHandler(void *cls, struct MHD_Connection *connection,
						const char *url,
						const char *method, const char *version,
						const char *upload_data,
						size_t *upload_data_size, void **con_cls);

void daemonMain(pid_t pid);
void printUsage();
void signal_handler(int sig);
void controlThreadJob(unsigned int sleepSeconds);
bool initConfiguration(string path);

int main(int argc, char** argv) {
	int opt;
	bool demonize = false;

	// Setup signal handling before we start
	signal(SIGHUP, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);

	while ((opt = getopt(argc, argv, "hda:p:i:s:c:")) != -1) {
		switch (opt) {
		case 'h':
			printUsage();
			exit(EXIT_SUCCESS);
			break;
		case 'd':
			demonize = true;
			break;
		case 'a':
			bindAddr = strndup(optarg, 16);
			break;
		case 'p':
			httpPort = atoi(optarg);
			break;
		case 'i':
			directoryFolder = Lucene::StringUtils::toUnicode(optarg);
			break;
		case 's':
			threadSleepSeconds = atoi(optarg);
			break;
		case 'v':
			printf("%s\n", PAGLIAIO_VERSION);
			exit(EXIT_SUCCESS);
			break;
		case 'c':
			configurationFilePath = strndup(optarg, 1024); //FIXME basteranno 1024 caratteri?
			break;
		default:
			printUsage();
			exit(EXIT_FAILURE);
			break;
		}
	}

	pid_t pid = 0, sid;

	if (demonize) {
		/* Fork off the parent process */
		pid = fork();
		if (pid < 0) {
			exit(EXIT_FAILURE);
		}
		/* If we got a good PID, then
		   we can exit the parent process. */
		if (pid > 0) {
			openlog("pagliaio", LOG_ODELAY, LOG_DAEMON);
			std::stringstream ss;
			ss << "pid: " << pid;
			syslog(LOG_INFO, "%s", ss.str().c_str());

			exit(EXIT_SUCCESS);
		}

		/* Change the file mode mask */
		umask(S_IWGRP | S_IWOTH);

		/* Create a new SID for the child process */
		sid = setsid();
		if (sid < 0) {
			/* Log any failure */
			exit(EXIT_FAILURE);
		}

		/* Change the current working directory */
		if ((chdir("/")) < 0) {
			/* Log any failure here */
			exit(EXIT_FAILURE);
		}

		/* Close out the standard file descriptors */
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	daemonMain(pid);

	return 0;
}

void daemonMain(pid_t pid) {
	openlog("pagliaio", LOG_ODELAY, LOG_DAEMON);

	syslog(LOG_INFO, "starting...");

	xmlInitParser();

	//leggi il file di configurazione
	bool confOk = initConfiguration(boost::lexical_cast<string>(configurationFilePath));
	if (!confOk) {
		syslog(LOG_ERR, "Error reading configuration file: %s", configurationFilePath);
	}

	if (!Lucene::FileUtils::isDirectory(directoryFolder)) {
		//crea la directory degli indici
		bool created = Lucene::FileUtils::createDirectory(directoryFolder);
		if (!created) {
			syslog(LOG_ERR, "Can't create indexes directory: %ls", directoryFolder.c_str());
			exit(EXIT_FAILURE);
		}
	}

	//crea l'indexWriter e l'indexSearcher prima che altri thread li vogliano usare...
	Lucene::FSDirectoryPtr dir = Lucene::FSDirectory::open(directoryFolder); //FIXME capire se se serve un altro tipo di directory (ad es. su database) o questo va bene
	Lucene::StandardAnalyzerPtr analyzer = Lucene::newLucene<Lucene::StandardAnalyzer>(Lucene::LuceneVersion::LUCENE_CURRENT); //FIXME capire se se serve un altro analizzatore o questo va bene

	try {
		indexWriter = Lucene::newLucene<Lucene::IndexWriter>(dir, analyzer, Lucene::IndexWriter::MaxFieldLengthLIMITED);
	}
	catch (...){
		syslog(LOG_ERR, "Caught an unknown exception");
		exit(EXIT_FAILURE);
	}

	indexReader = Lucene::IndexReader::open(dir, true);
	indexSearcher = Lucene::newLucene<Lucene::IndexSearcher>(indexReader);

	//crea il thread di controllo per l'update dell'indexReader
	boost::thread controlThread(controlThreadJob, threadSleepSeconds);

	struct sockaddr_in serverAddr;
	memset (&serverAddr, 0, sizeof (struct sockaddr_in));

	serverAddr.sin_family = AF_INET;
	serverAddr.sin_port = htons(httpPort);
	int validAddr = inet_pton(AF_INET, bindAddr, &serverAddr.sin_addr);

	if (validAddr == 0) {
		indexWriter->close(true);
		indexSearcher->close();
		indexReader->close();

		syslog(LOG_ERR, "Invalid address %s", bindAddr);
		exit(EXIT_FAILURE);
	}

	MHD_Daemon *httpDaemon = MHD_start_daemon(MHD_USE_THREAD_PER_CONNECTION, httpPort, NULL, NULL, httpRequestsHandler, NULL, MHD_OPTION_SOCK_ADDR, &serverAddr, MHD_OPTION_END);

	if (httpDaemon == NULL) {
		indexWriter->close(true);
		indexSearcher->close();
		indexReader->close();

		syslog(LOG_ERR, "Can't create http daemon");
		exit(EXIT_FAILURE);
	}

	while(daemonRunning) {
		//syslog(LOG_DEBUG, "sono vivo."); //FIXME rimuovere
		sleep(300); //FIXME pensare meglio che time out mettere
	}

	indexWriter->close(true);
	indexSearcher->close();
	indexReader->close();

	xmlCleanupParser();

	MHD_stop_daemon(httpDaemon);

	syslog(LOG_INFO, "done.");
}

void signal_handler(int sig) {
	openlog("pagliaio", LOG_ODELAY, LOG_DAEMON);

    switch(sig) {
	case SIGHUP:
		syslog(LOG_WARNING, "Received SIGHUP signal.");
		daemonRunning = false;
		break;
	case SIGTERM:
		syslog(LOG_WARNING, "Received SIGTERM signal.");
		daemonRunning = false;
		break;
	default:
		syslog(LOG_WARNING, "Unhandled signal (%d) %s", sig, strsignal(sig));
		break;
    }
}

void controlThreadJob(unsigned int sleepSeconds) {
	while (true) {
		boost::this_thread::sleep(boost::posix_time::seconds(sleepSeconds));

		boost::upgrade_lock<boost::shared_mutex> lock(indexMutex);
		Lucene::IndexReaderPtr indexReaderUpdated = indexReader->reopen();
		if (indexReaderUpdated != indexReader) {
			  // get exclusive access
			  boost::upgrade_to_unique_lock<boost::shared_mutex> uniqueLock(lock);
			  indexReader->decRef();

			  indexReader = indexReaderUpdated;
			  indexSearcher = Lucene::newLucene<Lucene::IndexSearcher>(indexReader);
		}
	}
}

bool initConfiguration(std::string path) {
	xmlDocPtr doc; /* the resulting document tree */

	//xmlInitParser();

	xmlParserCtxtPtr ctxt = xmlNewParserCtxt();
	xmlSetGenericErrorFunc(ctxt, parseErrorFunc);
	doc = xmlCtxtReadFile(ctxt, path.c_str(), NULL, XML_PARSE_NOBLANKS);
	if (doc == 0) {
		return false;
	}

	/* Create xpath evaluation context */
	xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
	if(xpathCtx == 0) {
	    xmlFreeParserCtxt(ctxt);
	    xmlFreeDoc(doc);
		return false;
	}

	xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression((xmlChar*)"/configuration/defaults/operator", xpathCtx);
	if (xpathObj != 0) {
		xmlElementPtr operatorEl = (xmlElementPtr)xpathObj->nodesetval->nodeTab[0];
		char* defaultOperator = (char*)xmlNodeListGetString(doc, operatorEl->xmlChildrenNode, 1);
		configuration.setDefaultOperator(boost::lexical_cast<string>(defaultOperator));
	}

	xpathObj = xmlXPathEvalExpression((xmlChar*)"/configuration/defaults/field", xpathCtx);
		if (xpathObj != 0) {
			xmlElementPtr operatorEl = (xmlElementPtr)xpathObj->nodesetval->nodeTab[0];
			char* defaultOperator = (char*)xmlNodeListGetString(doc, operatorEl->xmlChildrenNode, 1);
			configuration.setDefaultField(boost::lexical_cast<string>(defaultOperator));
		}

	xpathObj = xmlXPathEvalExpression((xmlChar*)"/configuration/defaults/indextype", xpathCtx);
	if (xpathObj != 0) {
		xmlElementPtr indexTypeEl = (xmlElementPtr)xpathObj->nodesetval->nodeTab[0];
		char* defaultIndexType = (char*)xmlNodeListGetString(doc, indexTypeEl->xmlChildrenNode, 1);
		configuration.setDefaultIndexType((boost::lexical_cast<string>(defaultIndexType)));
	}

	xpathObj = xmlXPathEvalExpression((xmlChar*)"/configuration/defaults/stored", xpathCtx);
	if (xpathObj != 0) {
		xmlElementPtr storingEl = (xmlElementPtr)xpathObj->nodesetval->nodeTab[0];
		char* storing = (char*)xmlNodeListGetString(doc, storingEl->xmlChildrenNode, 1);
		configuration.setDefaultStoring((boost::lexical_cast<string>(storing)));
	}

	xpathObj = xmlXPathEvalExpression((xmlChar*)"/configuration/indices/index", xpathCtx);
	if (xpathObj != 0) {
		for (int i = 0; i < xpathObj->nodesetval->nodeNr; i++) {
			xmlNodePtr node = xpathObj->nodesetval->nodeTab[i];
			string name = boost::lexical_cast<std::string>(node->name);
			if (node->type == XML_ELEMENT_NODE && name == "index") {
				xmlElementPtr indexEl = (xmlElementPtr) node;

				//get index path and clean it from spaces;
				string index =  string((char*)xmlNodeListGetString(doc, indexEl->xmlChildrenNode, 1));
				index = trim(index);

				char* storedStr = (char*)"";
				char* type = (char*)"";

				xmlAttributePtr attribute = indexEl->attributes;
				while(attribute != 0) {
					string attrName = boost::lexical_cast<string>(attribute->name);
					if (attrName == "stored") {
						storedStr = (char*)xmlNodeListGetString(doc, attribute->xmlChildrenNode, 1);
					}
					else if (attrName == "type" ) {
						type = (char*)xmlNodeListGetString(doc, attribute->xmlChildrenNode, 1);
					}

					attribute = (xmlAttributePtr)((xmlNodePtr)attribute->next);
				}

				//FIXME si può fare meglio?
				bool stored = false;
				if (strncmp(storedStr, "true", 4) || strncmp(storedStr, "yes", 3))
					stored = true;

				configuration.addIndex(index, stored, type);
			}
		}
	}

	xmlFreeParserCtxt(ctxt);
	xmlFreeDoc(doc);
    xmlXPathFreeContext(xpathCtx);

    //xmlCleanupParser();

	return true;
}

#define ERROR_MSG_BUF_SIZE 256
void parseErrorFunc(void *ctx, const char *msg, ...) {
   char string[ERROR_MSG_BUF_SIZE];
   va_list arg_ptr;

   va_start(arg_ptr, msg);
   vsnprintf(string, ERROR_MSG_BUF_SIZE, msg, arg_ptr);
   va_end(arg_ptr);
   syslog(LOG_ERR, "%s", string);
   return;
}

void printUsage() {
	std::cout << "Non sai usare Pagliaio." << std::endl;
	std::cout << "Usage: pagliaio [OPTIONS]" << std::endl;
	std::cout << "-d			demonize process" << std::endl;
	std::cout << "-a address 	specifies pagliaio bind address (default 0.0.0.0)" << std::endl;
	std::cout << "-p port 		specifies pagliaio bind port (default 4200)" << std::endl;
	std::cout << "-i indexpath 	choose lucene++ index path (default /var/run/pagliaio)" << std::endl;
	std::cout << "-v 			print pagliaio version and exit" << std::endl;
	std::cout << "-s seconds 	choose pagliaio index updater slumber duration in seconds (default 5)" << std::endl;
	std::cout << "-c path 	    choose pagliaio configuration file (default /etc/pagliaio/pagliaio.xml)" << std::endl;
}
